const C3 = self.C3;

const BEHAVIOR_INFO = {
    id: "skymen_bounding_box",
    Acts: {
      "SetEnabled": {
            "forward": (inst) => inst._SetEnabled,
            
            "autoScriptInterface": true,
            },
"AddObjectToCustomList": {
            "forward": (inst) => inst._AddObjectToCustomList,
            
            "autoScriptInterface": false,
            },
"RemoveFromList": {
            "forward": (inst) => inst._RemoveFromList,
            
            "autoScriptInterface": false,
            },
"ClearList": {
            "forward": (inst) => inst._ClearList,
            
            "autoScriptInterface": true,
            },
"AddLayerToCustomList": {
            "forward": (inst) => inst._AddLayerToCustomList,
            
            "autoScriptInterface": false,
            },
"RemoveLayerFromList": {
            "forward": (inst) => inst._RemoveLayerFromList,
            
            "autoScriptInterface": false,
            },
"SetBoundingBoxMode": {
            "forward": (inst) => inst._SetBoundingBoxMode,
            
            "autoScriptInterface": false,
            },
"UpdateBoundingBox": {
            "forward": (inst) => inst._UpdateBoundingBox,
            
            "autoScriptInterface": true,
            }
    },
    Cnds: {
      "IsEnabled": {
            "forward": (inst) => inst._IsEnabled,
            
            "autoScriptInterface": true,
          }
    },
    Exps: {
      "Count": {
            "forward": (inst) => inst._Count,
            
            "autoScriptInterface": true,
          }
    },
  };

const camelCasedMap = new Map();

function camelCasify(str) {
  // If the string is already camelCased, return it
  if (camelCasedMap.has(str)) {
    return camelCasedMap.get(str);
  }
  // Replace any non-valid JavaScript identifier characters with spaces
  let cleanedStr = str.replace(/[^a-zA-Z0-9$_]/g, " ");

  // Split the string on spaces
  let words = cleanedStr.split(" ").filter(Boolean);

  // Capitalize the first letter of each word except for the first one
  for (let i = 1; i < words.length; i++) {
    words[i] = words[i].charAt(0).toUpperCase() + words[i].substring(1);
  }

  // Join the words back together
  let result = words.join("");

  // If the first character is a number, prepend an underscore
  if (!isNaN(parseInt(result.charAt(0)))) {
    result = "_" + result;
  }

  camelCasedMap.set(str, result);

  return result;
}

C3.Behaviors[BEHAVIOR_INFO.id] = class extends C3.SDKBehaviorBase {
  constructor(opts) {
    super(opts);
  }

  Release() {
    super.Release();
  }
};
const B_C = C3.Behaviors[BEHAVIOR_INFO.id];
B_C.Type = class extends C3.SDKBehaviorTypeBase {
  constructor(objectClass) {
    super(objectClass);
  }

  Release() {
    super.Release();
  }

  OnCreate() {}
};

//====== SCRIPT INTERFACE ======
const map = new WeakMap();

function getScriptInterface(parentClass, map) {
  return class extends parentClass {
    constructor() {
      super();
      map.set(this, parentClass._GetInitInst().GetSdkInstance());
      this.INCLUDE_CHILDREN = {
        NONE: 0,
        OWN: 1,
        ALL: 2,
      };
      this.BOUNDING_BOX_MODE = {
        CUSTOM_LIST: 0,
        OWN_CHILDREN: 1,
        ALL_CHILDREN: 2,
      };
    }

    addObjectToCustomList(objectClass, includeChildrenMode) {
      const that = map.get(this);
      const instances = objectClass
        .getPickedInstances()
        .map((inst) => that._runtime.GetInstanceByUID(inst.uid));
      that._DoAddInstancesToCustomList(instances, includeChildrenMode);
    }
    removeFromList(objectClass) {
      const that = map.get(this);
      const instances = objectClass
        .getPickedInstances()
        .map((inst) => that._runtime.GetInstanceByUID(inst.uid));
      that._DoRemoveInstancesFromCustomList(instances);
    }
    _sdkLayerFromScriptLayer(layer) {
      if (!layer) return null;
      const that = map.get(this);
      if (typeof layer === "object") {
        layer = layer.name;
      }
      return that._runtime.GetMainRunningLayout().GetLayer(layer);
    }
    _AddLayerToCustomList(layer, childrenMode, subLayerMode) {
      const that = map.get(this);
      that._AddLayerToCustomList(
        this._sdkLayerFromScriptLayer(layer),
        subLayerMode,
        childrenMode
      );
    }
    removeLayerFromList(layer) {
      const that = map.get(this);
      that._RemoveLayerFromList(this._sdkLayerFromScriptLayer(layer));
    }
    setBoundingBoxMode(boundingBoxMode) {
      const that = map.get(this);
      that._SetBoundingBoxMode(boundingBoxMode);
    }
  };
}


const scriptInterface = getScriptInterface(self.IBehaviorInstance, map);

// extend script interface with plugin actions
Object.keys(BEHAVIOR_INFO.Acts).forEach((key) => {
  const ace = BEHAVIOR_INFO.Acts[key];
  if (!ace.autoScriptInterface) return;
  if (ace.isAsync) {
    scriptInterface.prototype[camelCasify(key)] = async function (...args) {
      const sdkInst = map.get(this);
      await B_C.Acts[camelCasify(key)].call(sdkInst, ...args);
    };
  } else {
    scriptInterface.prototype[camelCasify(key)] = function (...args) {
      const sdkInst = map.get(this);
      B_C.Acts[camelCasify(key)].call(sdkInst, ...args);
    };
  }
});

const addonTriggers = [];

// extend script interface with plugin conditions
Object.keys(BEHAVIOR_INFO.Cnds).forEach((key) => {
  const ace = BEHAVIOR_INFO.Cnds[key];
  if (!ace.autoScriptInterface || ace.isStatic || ace.isLooping) return;
  if (ace.isTrigger) {
    scriptInterface.prototype[camelCasify(key)] = function (callback, ...args) {
      const callbackWrapper = () => {
        const sdkInst = map.get(this);
        if (B_C.Cnds[camelCasify(key)].call(sdkInst, ...args)) {
          callback();
        }
      };
      this.addEventListener(key, callbackWrapper, false);
      return () => this.removeEventListener(key, callbackWrapper, false);
    };
  } else {
    scriptInterface.prototype[camelCasify(key)] = function (...args) {
      const sdkInst = map.get(this);
      return B_C.Cnds[camelCasify(key)].call(sdkInst, ...args);
    };
  }
});

// extend script interface with plugin expressions
Object.keys(BEHAVIOR_INFO.Exps).forEach((key) => {
  const ace = BEHAVIOR_INFO.Exps[key];
  if (!ace.autoScriptInterface) return;
  scriptInterface.prototype[camelCasify(key)] = function (...args) {
    const sdkInst = map.get(this);
    return B_C.Exps[camelCasify(key)].call(sdkInst, ...args);
  };
});
//====== SCRIPT INTERFACE ======

//============ ACES ============
B_C.Acts = {};
B_C.Cnds = {};
B_C.Exps = {};
Object.keys(BEHAVIOR_INFO.Acts).forEach((key) => {
  const ace = BEHAVIOR_INFO.Acts[key];
  B_C.Acts[camelCasify(key)] = function (...args) {
    if (ace.forward) ace.forward(this).call(this, ...args);
    else if (ace.handler) ace.handler.call(this, ...args);
  };
});
Object.keys(BEHAVIOR_INFO.Cnds).forEach((key) => {
  const ace = BEHAVIOR_INFO.Cnds[key];
  B_C.Cnds[camelCasify(key)] = function (...args) {
    if (ace.forward) return ace.forward(this).call(this, ...args);
    if (ace.handler) return ace.handler.call(this, ...args);
  };
  if (ace.isTrigger && ace.autoScriptInterface) {
    addonTriggers.push({
      method: B_C.Cnds[camelCasify(key)],
      id: key,
    });
  }
});
Object.keys(BEHAVIOR_INFO.Exps).forEach((key) => {
  const ace = BEHAVIOR_INFO.Exps[key];
  B_C.Exps[camelCasify(key)] = function (...args) {
    if (ace.forward) return ace.forward(this).call(this, ...args);
    if (ace.handler) return ace.handler.call(this, ...args);
  };
});
//============ ACES ============

function getInstanceJs(parentClass, scriptInterface, addonTriggers, C3) {
  return class extends parentClass {
    constructor(inst, properties) {
      super(inst);

      this.mode = 0;
      this.enabled = true;
      this.objectMap = new Map();

      this.customList = [];
      if (properties) {
        this.mode = properties[0];
        this.enabled = properties[1];
      }
      if (this.enabled) this._StartTicking();
    }

    Release() {
      super.Release();
    }

    SaveToJson() {
      return {
        mode: this.mode,
        enabled: this.enabled,
        customList: this.customList.map((x) => {
          if (x.isLayer) {
            return {
              isLayer: true,
              layer: x.layer.GetSID(),
              instMode: x.instMode,
              layerMode: x.layerMode,
            };
          } else {
            return {
              isLayer: false,
              inst: x.inst.GetUID(),
              mode: x.mode,
            };
          }
        }),
      };
    }

    LoadFromJson(o) {
      this.mode = o.mode;
      this.enabled = o.enabled;
      this.customList = o.customList.map((x) => {
        if (x.isLayer) {
          return {
            isLayer: true,
            layer: this._inst
              .GetRuntime()
              .GetCurrentLayout()
              .GetLayerBySID(x.layer),
            instMode: x.instMode,
            layerMode: x.layerMode,
          };
        } else {
          return {
            isLayer: false,
            inst: this._inst.GetRuntime().GetInstanceByUID(x.inst),
            mode: x.mode,
          };
        }
      });
    }

    _GetObjectsFromInst(inst, mode, map, eachCallback) {
      map.set(inst.GetUID(), inst);
      if (eachCallback) eachCallback(inst);
      if (mode !== 0) {
        inst
          .GetWorldInfo()
          .GetChildren()
          .forEach((childWi) => {
            this._GetObjectsFromInst(
              childWi._inst,
              mode === 1 ? 0 : 2,
              map,
              eachCallback
            );
          });
      }
    }

    _GetObjectsFromLayer(layer, instMode, layerMode, map, eachCallback) {
      layer._GetInstances().forEach((inst) => {
        this._GetObjectsFromInst(inst, instMode, map, eachCallback);
      });
      if (layerMode !== 0) {
        layer.GetSubLayers().forEach((sub) => {
          this._GetObjectsFromLayer(
            sub,
            instMode,
            layerMode === 1 ? 0 : 2,
            map,
            eachCallback
          );
        });
      }
    }

    _GetAllObjects(eachCallback) {
      const objectMap = this.objectMap;
      objectMap.clear();

      this.customList.forEach((x) => {
        if (x.isLayer) {
          this._GetObjectsFromLayer(
            x.layer,
            x.instMode,
            x.layerMode,
            objectMap,
            eachCallback
          );
        } else {
          this._GetObjectsFromInst(x.inst, x.mode, objectMap, eachCallback);
        }
      });

      if (this.mode !== 0) {
        this._inst
          .GetWorldInfo()
          .GetChildren()
          .forEach((childWi) => {
            this._GetObjectsFromInst(
              childWi._inst,
              this.mode === 1 ? 0 : 2,
              objectMap,
              eachCallback
            );
          });
      }

      return Array.from(objectMap.values());
    }

    _UpdateBoundingBox() {
      let minX = Number.MAX_SAFE_INTEGER;
      let minY = Number.MAX_SAFE_INTEGER;
      let maxX = Number.MIN_SAFE_INTEGER;
      let maxY = Number.MIN_SAFE_INTEGER;
      let anyInst = false;

      this._GetAllObjects((inst) => {
        anyInst = true;
        const bbox = inst.GetWorldInfo().GetBoundingBox();
        minX = Math.min(minX, bbox.getLeft());
        minY = Math.min(minY, bbox.getTop());
        maxX = Math.max(maxX, bbox.getRight());
        maxY = Math.max(maxY, bbox.getBottom());
      });

      if (!anyInst) {
        minX = 0;
        minY = 0;
        maxX = 0;
        maxY = 0;
      }

      const wi = this._inst.GetWorldInfo();
      const ox = wi.GetOriginX();
      const oy = wi.GetOriginY();

      const width = maxX - minX;
      const height = maxY - minY;

      const wiArr = [];
      const children = [...wi.GetChildren()];
      children.forEach((childWi) => {
        wiArr.push({
          wi: childWi,
          transformX: childWi.GetTransformWithParentX(),
          transformY: childWi.GetTransformWithParentY(),
          transformWidth: childWi.GetTransformWithParentWidth(),
          transformHeight: childWi.GetTransformWithParentHeight(),
          transformAngle: childWi.GetTransformWithParentAngle(),
          transformZElevation: childWi.GetTransformWithParentZElevation(),
          transformOpacity: childWi.GetTransformWithParentOpacity(),
          transformVisibility: childWi.GetTransformWithParentVisibility(),
          destroyWithParent: childWi.GetDestroyWithParent(),
        });
        wi.RemoveChild(childWi);
      });
      wi.SetSize(width, height);
      wi.SetXY(minX + ox * width, minY + oy * height);
      wi.SetBboxChanged();
      wiArr.forEach((x) => {
        wi.AddChild(x.wi, x);
      });
    }

    Tick() {
      this._UpdateBoundingBox();
    }

    Trigger(method) {
      super.Trigger(method);
      const addonTrigger = addonTriggers.find((x) => x.method === method);
      if (addonTrigger) {
        this.GetScriptInterface().dispatchEvent(new C3.Event(addonTrigger.id));
      }
    }

    GetScriptInterfaceClass() {
      return scriptInterface;
    }

    // ACTS
    _SetEnabled(e) {
      this.enabled = e;
      if (this.enabled) this._StartTicking();
      else this._StopTicking();
    }
    _DoAddInstancesToCustomList(instances, mode) {
      instances.forEach((inst) => {
        this.customList.push({ isLayer: false, inst, mode });
      });
    }
    _AddObjectToCustomList(object, mode) {
      this._DoAddInstancesToCustomList(
        object.GetCurrentSol().GetInstances(),
        mode
      );
    }
    _DoRemoveInstancesFromCustomList(insts) {
      this.customList = this.customList.filter((x) => !insts.includes(x.inst));
    }
    _RemoveFromList(object) {
      this._DoRemoveInstancesFromCustomList(
        object.GetCurrentSol().GetInstances()
      );
    }
    _ClearList() {
      this.customList = [];
    }
    _AddLayerToCustomList(layer, instMode, layerMode) {
      this.customList.push({ isLayer: true, layer, instMode, layerMode });
    }
    _RemoveLayerFromList(layer) {
      this.customList = this.customList.filter((x) => x.layer !== layer);
    }
    _SetBoundingBoxMode(m) {
      this.mode = m;
    }

    // CNDS
    _IsEnabled() {
      return this.enabled;
    }

    // EXPS
    _Count() {
      return this._GetAllObjects().length;
    }
  };
}


B_C.Instance = getInstanceJs(
  C3.SDKBehaviorInstanceBase,
  scriptInterface,
  addonTriggers,
  C3
);
